@@ -30,7 +30,6 @@
 
 #include <stdlib.h>
 #include <sys/types.h>
-#include <dirent.h>
 #include <string.h>
 #include <libxml/parser.h>
 #include <libxml/parserInternals.h>
@@ -150,14 +149,21 @@ parse_oaf_server_attrs (ParseInfo      *info,
         error = od_validate (iid, type, location);
         
         if (error != NULL) {
-                /* FIXME: should syslog */
-                g_print ("%s\n", error);
+                g_warning ("%s", error);
                 
                 g_free (error);
 
                 return;
         }
 
+#ifdef G_OS_WIN32
+        /* Possibly replace configure-time shlib or exe location
+         * with the actual installed one.
+         */
+        if (!strcmp (type, "exe") || !strcmp (type, "shlib"))
+                location = server_win32_replace_prefix (location);
+#endif
+
         /* Now create the ServerInfo object */
         info->cur_server = g_new0 (Bonobo_ServerInfo, 1);
 
@@ -167,6 +173,11 @@ parse_oaf_server_attrs (ParseInfo      *info,
         info->cur_server->hostname = CORBA_string_dup (info->host);
         info->cur_server->username = CORBA_string_dup (g_get_user_name ());
         info->cur_server->domain = CORBA_string_dup ("unused");
+
+#ifdef G_OS_WIN32
+        if (!strcmp (type, "exe") || !strcmp (type, "shlib"))
+                g_free ((char *) location);
+#endif
 }
 
 static GHashTable *interesting_locales = NULL;
@@ -174,16 +185,17 @@ static GHashTable *interesting_locales = NULL;
 void
 add_initial_locales (void)
 {
-        const GList *l;
+        const char * const * langs;
+        int i;
 
         if (!interesting_locales)
                 interesting_locales = g_hash_table_new (
                         g_str_hash, g_str_equal);
 
-	for (l = bonobo_activation_i18n_get_language_list (NULL);
-             l; l = l->next)
+        langs = g_get_language_names ();
+	for (i = 0; langs[i] != NULL; i++)
 		g_hash_table_insert (interesting_locales,
-                                     g_strdup ((char *) l->data),
+                                     g_strdup (langs[i]),
                                      GUINT_TO_POINTER (1));
 }
 
@@ -279,11 +291,11 @@ parse_oaf_attribute (ParseInfo     *info,
                 return;
         
         if (name[0] == '_')
-                g_error ("%s is an invalid property name "
-                         "- property names beginning with '_' are reserved",
-                         name);
+                g_critical ("%s is an invalid property name "
+                            "- property names beginning with '_' are reserved",
+                            name);
         
-        info->cur_prop = g_new0 (Bonobo_ActivationProperty, 1);
+        info->cur_prop = ORBit_small_alloc (TC_Bonobo_ActivationProperty);
         info->cur_prop->name = CORBA_string_dup (name);
 
         if (g_ascii_strcasecmp (type, "stringv") == 0) {
@@ -454,11 +466,13 @@ od_EndElement (ParseInfo     *info,
                         len = g_list_length (info->cur_props);
 
                         info->cur_server->props._length = len;
-                        info->cur_server->props._buffer = g_new0 (Bonobo_ActivationProperty, len);
+                        info->cur_server->props._buffer =
+                                CORBA_sequence_Bonobo_ActivationProperty_allocbuf (len);
 
                         for (i = 0, p = g_list_reverse (info->cur_props); p; p = p->next, i++) {
-                                info->cur_server->props._buffer[i] = *((Bonobo_ActivationProperty *)p->data);
-                                g_free (p->data);
+                                Bonobo_ActivationProperty_copy (&info->cur_server->props._buffer[i],
+                                                                (Bonobo_ActivationProperty *) p->data);
+                                CORBA_free (p->data);
                         }
                         g_list_free (info->cur_props);
                         info->cur_props = NULL;
@@ -551,19 +565,14 @@ static xmlSAXHandler od_SAXParser = {
 };
 
 static void
-od_load_file (const char *file,
-              GSList    **entries,
-              const char *host)
+od_load_context (xmlParserCtxt *ctxt,
+                 GSList    **entries,
+                 const char *host)
 {
         ParseInfo *info;
 	xmlSAXHandlerPtr oldsax;
-        xmlParserCtxt *ctxt;
         int ret = 0;
 
-        ctxt = xmlCreateFileParserCtxt (file);
-        if (!ctxt)
-                return;
-        
         info = parse_info_new ();
         info->host = host;
         info->entries = entries;
@@ -585,16 +594,43 @@ od_load_file (const char *file,
                         ret = -1;
         }
         ctxt->sax = oldsax;
-        xmlFreeParserCtxt (ctxt);
 
         parse_info_free (info);
 
         if (ret < 0) {
-                /* FIXME: syslog the error */
+                g_warning (_("Could not parse badly formed XML document %s"), ctxt->input->filename);
                 return;
         }
 }
 
+static void
+od_load_file (const char *file,
+              GSList    **entries,
+              const char *host)
+{
+        xmlParserCtxt *ctxt;
+
+        ctxt = xmlCreateFileParserCtxt (file);
+        if (!ctxt)
+                return;
+        od_load_context (ctxt, entries, host);
+        xmlFreeParserCtxt (ctxt);
+}
+
+void
+bonobo_parse_server_info_memory (const char *server_info,
+                                 GSList    **entries,
+                                 const char *host)
+{
+        xmlParserCtxt *ctxt;
+
+        ctxt = xmlCreateMemoryParserCtxt (server_info, strlen (server_info));
+        if (!ctxt)
+                return;
+        od_load_context (ctxt, entries, host);
+        xmlFreeParserCtxt (ctxt);
+}
+
 static gboolean
 od_filename_has_extension (const char *filename,
                            const char *extension)
@@ -611,25 +647,23 @@ od_load_directory (const char *directory,
                    GSList    **entries,
                    const char *host)
 {
-	DIR *directory_handle;
-	struct dirent *directory_entry;
+	GDir *directory_handle;
+	const char *directory_entry;
         char *pathname;
 
-        
-        /* FIXME: Should be a syslog message. */
-        /* g_print (_("Trying dir %s\n"), directory); */
+        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, _("Trying dir %s"), directory);
 
-        directory_handle = opendir (directory);
+        directory_handle = g_dir_open (directory, 0, NULL);
 
         if (directory_handle == NULL) {
                 /* FIXME */
                 return;
         }
         
-        for (directory_entry = readdir (directory_handle);
+        for (directory_entry = g_dir_read_name (directory_handle);
              directory_entry != NULL;
-             directory_entry = readdir (directory_handle)) {
-                pathname = g_strdup_printf ("%s/%s", directory, directory_entry->d_name);
+             directory_entry = g_dir_read_name (directory_handle)) {
+                pathname = g_build_filename (directory, directory_entry, NULL);
 
                 if (od_filename_has_extension (pathname, ".server")) {
                         od_load_file (pathname, entries, host);
@@ -638,15 +672,16 @@ od_load_directory (const char *directory,
 		g_free (pathname);
         }
 
-        closedir (directory_handle);
+        g_dir_close (directory_handle);
 }
 
 
 void
-bonobo_server_info_load (char **directories,
-                         Bonobo_ServerInfoList   *servers,
-                         GHashTable **iid_to_server_info_map,
-                         const char *host)
+bonobo_server_info_load (char                  **directories,
+                         Bonobo_ServerInfoList  *servers,
+                         GPtrArray const        *runtime_servers,
+                         GHashTable            **iid_to_server_info_map,
+                         const char             *host)
 {
 	GSList *entries;
         int length;
@@ -671,8 +706,10 @@ bonobo_server_info_load (char **directories,
 	/* Now convert 'entries' into something that the server can store and pass back */
 	length = g_slist_length (entries);
 
-	servers->_buffer = CORBA_sequence_Bonobo_ServerInfo_allocbuf (length);
-        servers->_length = length;
+	servers->_buffer = CORBA_sequence_Bonobo_ServerInfo_allocbuf
+                (length + runtime_servers->len);
+	servers->_length = length + runtime_servers->len;
+        servers->_maximum = servers->_length;
 
 	for (j = 0, p = entries; j < length; j++, p = p->next) {
 		memcpy (&servers->_buffer[j], p->data, sizeof (Bonobo_ServerInfo));
@@ -680,6 +717,14 @@ bonobo_server_info_load (char **directories,
                                      servers->_buffer[j].iid,
                                      &servers->_buffer[j]);
 	}
+          /* append information of runtime-defined servers  */
+	for (j = 0; j < runtime_servers->len; j++) {
+                servers->_buffer[length + j] = *(Bonobo_ServerInfo *)
+                        g_ptr_array_index (runtime_servers, j);
+		g_hash_table_insert (*iid_to_server_info_map,
+                                     servers->_buffer[length + j].iid,
+                                     &servers->_buffer[length + j]);
+	}
 
         g_slist_foreach (entries, (GFunc) g_free, NULL);
         g_slist_free (entries);
